/* wptMDSumDlg.cpp - Dialog to show hash values for files
 *	Copyright (C) 2003, 2005, 2006 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <stdio.h>

#include "resource.h"
#include "wptTypes.h"
#include "wptW32API.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptContext.h"
#include "wptNLS.h"
#include "wptErrors.h"


/* maximum hash size in octets (sha256=32) */
#define MAX_HASHSIZE 32

/* Symbolic column IDs. */
enum md_col_t {COL_MD=0, COL_NAME};


/* A model which represents the contents of the list view. */
struct hashlist_model_s {
    struct hashlist_model_s *next;
    char *name;
    BYTE *md;
    DWORD mdlen;
};
typedef struct hashlist_model_s *hashlist_model_t;


/* Release the file list @hm. */
static void
hashmodel_release (hashlist_model_t hm)
{
    hashlist_model_t t;

    while (hm) {
	t = hm->next;
	free_if_alloc (hm->name);
	free_if_alloc (hm->md);
	free_if_alloc (hm);
	hm = t;
    }
}


/* Add a new file with the name @name to the list @hm. */
static hashlist_model_t
hashmodel_add_file (hashlist_model_t *hm, const char *name, 
		    const BYTE *mdbuf, size_t mdlen)
{
    hashlist_model_t t, n;

    t = new hashlist_model_s;
    if (!t)
	BUG (NULL);
    memset (t, 0, sizeof *t);
    t->name = m_strdup (name);
    t->mdlen = mdlen;
    t->md = new BYTE[mdlen];
    if (!t->md)
	BUG (NULL);
    memcpy (t->md, mdbuf, mdlen);
    if (!*hm)
	*hm = t;
    else {
	for (n = *hm; n->next; n=n->next)
	    ;
	n->next = t;
    }
    return t;
}


/* Return a printable digest of the buffer @mdbuf. */
static const char*
printable_digest (BYTE *mdbuf, size_t n)
{
    static char mdasc[4*MAX_HASHSIZE];
    size_t i;

    for (i = 0; i < n; i++)
	sprintf (mdasc+2*i, "%02X", mdbuf[i]);
    return mdasc;
}


static const char*
id2algo (gpgme_hash_algo_t mdalgo)
{
    switch (mdalgo) {
    case GPGME_MD_MD5:	    return "MD5";
    case GPGME_MD_SHA1:	    return "SHA1";
    case GPGME_MD_RMD160:   return "RMD160";
    case GPGME_MD_SHA256:   return "SHA256";
    case GPGME_MD_SHA384:   return "SHA384";
    case GPGME_MD_SHA512:   return "SHA512";
    default:		    break;
    }
    return "";
}


/* Hash the selected file from the FM listview control in @md.
   Add the results to the listview @lv. */
static void
hash_selected_files (md_file_s *md, listview_ctrl_t lv, hashlist_model_t *r_fl)
{
    hashlist_model_t item;
    BYTE mdbuf[MAX_HASHSIZE];
    char fname[256+1];
    size_t n;
    int i;

    for (i = 0; i < listview_count_items (md->lv, 0); i++) {
	if (!listview_get_item_state (md->lv, i))
	    continue;
	listview_get_item_text (md->lv, i, 1, fname, sizeof (fname)-1);
	if (!gpg_md_hash_file (md->mdalgo, fname, mdbuf, &n)) {
	    item = hashmodel_add_file (r_fl, fname, mdbuf, n);
	    listview_add_item2 (lv, "", (void*)item);
	    listview_add_sub_item (lv, 0, COL_MD, printable_digest (mdbuf, n));
	    listview_add_sub_item (lv, 0, COL_NAME, fname);
	}	
    }
}


/* Return 1 if there is a tool available to check the file.
   (sha1sum, md5sum). */
static int
tool_avail (gpgme_hash_algo_t mdalgo)
{
    if (mdalgo == GPGME_MD_SHA1 || mdalgo == GPGME_MD_MD5)
	return 1;
    return 0;
}


/* Sorting callback. */
static int CALLBACK
sort_cb (LPARAM first, LPARAM second, LPARAM param)
{
    hashlist_model_t a, b;
    int sortby = (int)param;
    int cmpres = 0;

    a = (hashlist_model_t)first;
    b = (hashlist_model_t)second;
    if (sortby == COL_NAME)
	cmpres = stricmp (a->name, b->name);
    else
	cmpres = memcmp (a->md, b->md, a->mdlen);
    return cmpres;
}


/* Dialog box procedure to show and calculate file digests. */
BOOL CALLBACK
mdsum_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static listview_ctrl_t lv;
    static struct md_file_s *md;
    static hashlist_model_t hm = NULL;
    struct listview_column_s cols[] = {
        {0, 264, (char *)_("Digest")},
        {1, 160, (char *)_("Name")},
        {0, 0, 0}
    };
    gpgme_data_t sumlist = NULL;
    const char *name, *algname;
    char fname[MAX_PATH+64], mdasc[MAX_HASHSIZE*4];
    int i;

    switch (msg) {
    case WM_INITDIALOG:
	md = (md_file_s *)lparam;
	if (!md)
	    BUG (NULL);
	listview_new (&lv, GetDlgItem (dlg, IDC_MDSUM_LIST));
	for (i = 0; i < cols[i].width; i++)
	    listview_add_column (lv, &cols[i]);
	hash_selected_files (md, lv, &hm);
	SetDlgItemText (dlg, IDC_MDSUM_COPY, _("&Save..."));
	SetDlgItemText (dlg, IDOK, _("&Close"));
	SetDlgItemText (dlg, IDC_MDSUM_TOCLIP, _("Save to clipboard"));
	SetWindowText (dlg, _("Print Message Digest"));
	SetForegroundWindow (dlg);
	break;

    case WM_DESTROY:
	if (lv) {
	    listview_release (lv);
	    lv = NULL;
	}
	if (hm) {
	    hashmodel_release (hm);
	    hm = NULL;
	}
	break;

    case WM_NOTIFY:
	if (((NMHDR *)lparam)->code == LVN_COLUMNCLICK) {
	    NMLISTVIEW *nft = (LPNMLISTVIEW) lparam;
	    ListView_SortItems (lv->ctrl, sort_cb, nft->iSubItem);
	}
	break;

    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case IDOK:
	    EndDialog (dlg, TRUE);
	    break;

	case IDCANCEL:
	    EndDialog (dlg, FALSE);
	    break;

	case IDC_MDSUM_COPY:
	    algname = id2algo ((gpgme_hash_algo_t)md->mdalgo);
	    if (gpgme_data_new (&sumlist))
		BUG(0);
	    if (!tool_avail ((gpgme_hash_algo_t)md->mdalgo)) {
		const char *s;

		s = "#warning '";
		gpgme_data_write (sumlist, s, strlen (s));
		gpgme_data_write (sumlist, algname, strlen (algname));
		s = " ' sum is not yet available\r\n";
		gpgme_data_write (sumlist, s, strlen (s));
	    }
	    for (i = 0; i < listview_count_items (lv, 0); i++) {
		listview_get_item_text (lv, i, COL_MD, mdasc, DIM (mdasc)-1);
		listview_get_item_text (lv, i, COL_NAME, fname, DIM (fname)-1);
		
		gpgme_data_write (sumlist, mdasc, strlen (mdasc));
		gpgme_data_write (sumlist, " ", 1);
		gpgme_data_write (sumlist, fname, strlen (fname));
		gpgme_data_write (sumlist, "\r\n", 2);
	    }
	    if (IsDlgButtonChecked (dlg, IDC_MDSUM_TOCLIP)) {
		gpg_data_release_and_set_clipboard (sumlist, 0);
		break;
	    }
	    _snprintf (fname, sizeof (fname)-1, "%s_sums.txt", algname);
	    name = get_filesave_dlg (dlg, _("Select File to Save Checksums"), 
				     NULL, fname);
	    if (name && *name) {
		gpgme_error_t err;

		err = gpg_data_release_and_set_file (sumlist, name);
		if (!err) {
		    log_box (_("File Manager"), MB_OK, 
			     _("Checksums successfully saved in '%s'"), name);
		    sumlist = NULL;
		}
		else
		    msg_box (dlg, gpgme_strerror (err), _("File Manager"), MB_ERR);
	    }
	    if (sumlist)
		gpgme_data_release (sumlist);
	    break;
	}
	break;
    }
    return FALSE;
}
